/* ---------------------------------------------

	[ oledsub.c ]

	OLED`֘ATu[`
	For RGS15128128FH029
	(c) 2010, By Y.Kyotani

--------------------------------------------- */

#include	<avr/io.h>
#include	<avr/pgmspace.h>

#include	"integer.h"
#include	"timer168.h"
#include	"oledsub.h"

/* ------------------------------------ PORTC */
#define		_CS	0
#define		_RES	1
#define		D_C	2
#define		_WR	3
#define		_RD	4
#define		VCC	5

/* ---------------------------------- STR CMD */
#define		CMD_ENDJOB	0xfffe
#define		CMD_MSTIMER	0xffff
#define		CMD_OL		0x8000

/* ------------------------------ SCREEN SIZE */
#define SCREENGX     128
#define SCREENGY     128


/* \̒` */
typedef struct					// LN^Agr[g
	{
	WORD	chrx, chry;
	WORD	gx, gy, gdx, gdy;
	WORD	fgcolor, bgcolor;
	}
	ATTR_OLEDCHR;

/* O[oϐ */
ATTR_OLEDCHR	attr;				// LN^Agr[g

/* ASCIItHg(8x8Dot) */
extern const prog_uchar ankfont[];
extern const prog_uchar exfont[];


/* ---------------------------------------------

	[J֐vg^Cv

--------------------------------------------- */
void OLEDCmdOne(BYTE cmd);
void OLEDDataOne(BYTE data);
void OLEDStrOut(const prog_uint16_t *outstr);
void OLEDFillBoxA(void);
void OLEDPutChrP(const prog_uchar *p1, const prog_uchar *p2);


/* ---------------------------------------------

	f[^o͊֐

--------------------------------------------- */

void OLEDCmdOne(BYTE cmd)
	{
	PORTC &= ~(1<<D_C);	// D/C#
	PORTC &= ~(1<<_CS);	// CS#
	PORTD = cmd;		// CMDݒ
	PORTC &= ~(1<<_WR);	// WR#
	PORTC |= (1<<_WR);	// WR#
	PORTC |= (1<<_CS);	// CS#
	PORTC |= (1<<D_C);	// D/C#
	}

void OLEDDataOne(BYTE data)
	{
	PORTC &= ~(1<<_CS);	// CS#
	PORTD = data;		// Dataݒ
	PORTC &= ~(1<<_WR);	// WR#
	PORTC |= (1<<_WR);	// WR#
	PORTC |= (1<<_CS);	// CS#
	}

void OLEDStrOut(const prog_uint16_t *outstr)
	{
	WORD	data;
	
	for	( ; ; )
		{
		switch	(data= pgm_read_word_near(outstr++))
			{
			case	CMD_ENDJOB:
				return;
			case	CMD_MSTIMER:
				data= pgm_read_word_near(outstr++);
				MSTimer	(data);
				continue;
			default:
				if	(data & CMD_OL)
					OLEDCmdOne	((BYTE)(data & 0xff));
				else
					OLEDDataOne	((BYTE)data);
			}
		}
	}

/* ---------------------------------------------

	OLED֐

--------------------------------------------- */
const prog_uint16_t InitData[]=
	{
	CMD_OL | 0xfd,				// UnLock
		 0xb3,
	CMD_OL | 0xca,				// SetMUX
		 0x7f,				// ->128
	CMD_OL | 0x36,				// MADCTL
		 0x80,
		 0x01,
	CMD_OL | 0xba,				// Set Contrast(100cd)
		 0x53,				// ColorA: Red
	CMD_OL | 0xbb,
		 0x67,				// ColorB: Green
	CMD_OL | 0xbc,
		 0x64,				// ColorC: Blue
	CMD_OL | 0xbe,				// Gamma Lookup Table
		 1,				// ColorA: Red0
		 1,				// ColorA: Red1
		 2,				// ColorA: Red2
		 4,				// ColorA: Red3
		 6,				// ColorA: Red4
		 8,				// ColorA: Red5
		 10,				// ColorA: Red6
		 12,				// ColorA: Red7
		 14,				// ColorA: Red8
		 17,				// ColorA: Red9
		 20,				// ColorA: Red10
		 23,				// ColorA: Red11
		 27,				// ColorA: Red12
		 33,				// ColorA: Red13
		 35,				// ColorA: Red14
		 37,				// ColorA: Red15
		 41,				// ColorA: Red16
		 47,				// ColorA: Red17
		 51,				// ColorA: Red18
		 57,				// ColorA: Red19
		 65,				// ColorA: Red20
		 70,				// ColorA: Red21
		 75,				// ColorA: Red22
		 81,				// ColorA: Red23
		 86,				// ColorA: Red24
		 91,				// ColorA: Red25
		 97,				// ColorA: Red26
		 101,				// ColorA: Red27
		 113,				// ColorA: Red28
		 119,				// ColorA: Red29
		 123,				// ColorA: Red30
		 127,				// ColorA: Red31
		 1,				// ColorB: Green0
		 1,				// ColorB: Green1
		 2,				// ColorB: Green2
		 4,				// ColorB: Green3
		 6,				// ColorB: Green4
		 8,				// ColorB: Green5
		 10,				// ColorB: Green6
		 12,				// ColorB: Green7
		 14,				// ColorB: Green8
		 17,				// ColorB: Green9
		 20,				// ColorB: Green10
		 23,				// ColorB: Green11
		 27,				// ColorB: Green12
		 33,				// ColorB: Green13
		 35,				// ColorB: Green14
		 37,				// ColorB: Green15
		 41,				// ColorB: Green16
		 47,				// ColorB: Green17
		 51,				// ColorB: Green18
		 57,				// ColorB: Green19
		 65,				// ColorB: Green20
		 70,				// ColorB: Green21
		 75,				// ColorB: Green22
		 81,				// ColorB: Green23
		 86,				// ColorB: Green24
		 91,				// ColorB: Green25
		 97,				// ColorB: Green26
		 101,				// ColorB: Green27
		 113,				// ColorB: Green28
		 119,				// ColorB: Green29
		 123,				// ColorB: Green30
		 127,				// ColorB: Green31
		 1,				// ColorC: Blue0
		 2,				// ColorC: Blue1
		 3,				// ColorC: Blue2
		 4,				// ColorC: Blue3
		 6,				// ColorC: Blue4
		 8,				// ColorC: Blue5
		 10,				// ColorC: Blue6
		 12,				// ColorC: Blue7
		 14,				// ColorC: Blue8
		 17,				// ColorC: Blue9
		 20,				// ColorC: Blue10
		 24,				// ColorC: Blue11
		 28,				// ColorC: Blue12
		 33,				// ColorC: Blue13
		 36,				// ColorC: Blue14
		 39,				// ColorC: Blue15
		 43,				// ColorC: Blue16
		 49,				// ColorC: Blue17
		 51,				// ColorC: Blue18
		 54,				// ColorC: Blue19
		 57,				// ColorC: Blue20
		 65,				// ColorC: Blue21
		 70,				// ColorC: Blue22
		 75,				// ColorC: Blue23
		 81,				// ColorC: Blue24
		 86,				// ColorC: Blue25
		 91,				// ColorC: Blue26
		 97,				// ColorC: Blue27
		 101,				// ColorC: Blue28
		 113,				// ColorC: Blue29
		 119,				// ColorC: Blue30
		 127,				// ColorC: Blue31
	CMD_OL | 0x3a,				// Interface Pixel Format
		 0x05,				// 65K color
	CMD_OL | 0x28,				// All Pixels OFF
	CMD_OL | 0x11,				// Sleep Out
	CMD_ENDJOB
	};

void OLEDInit(void)
	{
	MSTimer		(50);			// 50msWait
	PORTC = 0b00011111;			// VCC:0, CS#:1, RES#:1, DC#:1, WR#:1, RD#:1
	MSTimer		(10);			// 10msWait
	PORTC &= ~(1<<_RES);			// RES#
	MSTimer		(10);			// 10msWait
	PORTC |= (1<<_RES);			// RES#
	MSTimer		(200);			// 200msWait
	PORTC |= (1<<VCC);			// VCC
	MSTimer		(300);			// 300msWait
	
	OLEDStrOut	(InitData);
	}

/* ---------------------------------------------

	fBXvC ON/OFF؂ւ

--------------------------------------------- */

void OLEDDispON(void)
	{
	OLEDCmdOne	(0x29);			// Disable All Pixels ON/OFF
	}

void OLEDDispOFF(void)
	{
	OLEDCmdOne	(0x28);			// All Pixels OFF
	}

/* ---------------------------------------------

	EBhEGAݒ肷

--------------------------------------------- */

int SetWindowArea(int dx, int dy, int ofstx, int ofsty)
	{
	int	sx, sy, res;
	
	if	(dx > SCREENGX || dy > SCREENGY)
		return	2;
	
	res= 0;
	sx= (SCREENGX - dx) / 2 + ofstx;
	if	(sx < 0)		{ sx=0; res=1; }
	else if	(sx > SCREENGX-dx)	{ sx=SCREENGX-dx; res=1; }
	sy= (SCREENGY - dy) / 2 + ofsty;
	if	(sy < 0)		{ sy=0; res=1; }
	else if	(sy > SCREENGY-dy)	{ sy=SCREENGY-dy; res=1; }
	
	SetWindowParam	((WORD)sx, (WORD)sy, (WORD)(sx+dx-1), (WORD)(sy+dy-1));
	
	return	res;
	}

void SetWindowParam(WORD x1, WORD y1, WORD x2, WORD y2)
	{
	OLEDCmdOne	(0x2a);			// Set Column Addres
	OLEDDataOne	((BYTE)x1);
	OLEDDataOne	((BYTE)x2);
	OLEDCmdOne	(0x2b);			// Set Row Addres
	OLEDDataOne	((BYTE)y1);
	OLEDDataOne	((BYTE)y2);
	
	OLEDCmdOne	(0x2c);			// Memory Write
	}

void InitWindowArea(void)
	{
	SetWindowParam	(0, 0, SCREENGX-1, SCREENGY-1);
	}


/* ---------------------------------------------

	ʑŜEwFŖ߂

--------------------------------------------- */

void GCls(WORD col)
	{
	InitWindowArea	();
	
	OLEDFill	((DWORD)SCREENGX * (DWORD)SCREENGY, col);
	}


/* ---------------------------------------------

	OLEDwFŖ߂

--------------------------------------------- */

void OLEDFill(DWORD len, WORD col)
	{
	/* Data */
	PORTC &= ~(1<<_CS);			// CS#

	while	(len >= 8L)
		{
		// 1
		PORTD = (BYTE)(col >> 8);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		PORTD = (BYTE)(col & 0xff);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		// 2
		PORTD = (BYTE)(col >> 8);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		PORTD = (BYTE)(col & 0xff);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		// 3
		PORTD = (BYTE)(col >> 8);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		PORTD = (BYTE)(col & 0xff);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		// 4
		PORTD = (BYTE)(col >> 8);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		PORTD = (BYTE)(col & 0xff);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		len -= 4L;
		}

	while	(len)
		{
		// 1
		PORTD = (BYTE)(col >> 8);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		PORTD = (BYTE)(col & 0xff);		// ʃoCg
		PORTC &= ~(1<<_WR);			// WR
		PORTC |= (1<<_WR);			// WR
		
		len--;
		}


	PORTC |= (1<<_CS);			// CS#
	}

/* ---------------------------------------------

	obt@Oꂽf[^]

--------------------------------------------- */
void OLEDTrans(BYTE *buff, WORD len)
	{
	/* Data */
	PORTC &= ~(1<<_CS);			// CS#

	while	(len >= 4)
		{
		// 1
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		// 2
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		// 3
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		// 4
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#

		len -= 4;
		}

	while	(len)
		{
		// 1
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		
		PORTD = *buff++;
		PORTC &= ~(1<<_WR);		// WR#
		PORTC |= (1<<_WR);		// WR#
		
		len--;
		}

	PORTC |= (1<<_CS);			// CS#
	}

/* ---------------------------------------------

	ςMMCf[^_CNg]
	len: 1Dot(2byte)Pʂ̒

--------------------------------------------- */
void MMC2OLEDDirect(WORD len)
	{
	/* Data */
	PORTC &= ~(1<<_CS);			// CS#

	SPDR = 0xFF;				// SPIsX^[g
	while	(len >= 4)
		{
		// 1
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD= SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD = SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		// 2
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD= SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD = SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		// 3
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD= SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD = SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		// 4
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD= SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD = SPDR;
		if	((len-=4) > 0)	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		}

	while	(len)
		{
		// 1
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD= SPDR;	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		
		loop_until_bit_is_set(SPSR, SPIF);
		PORTD = SPDR;
		if	(--len > 0)	SPDR = 0xFF;
		PORTC &= ~(1<<_WR);			// WR#
		PORTC |= (1<<_WR);			// WR#
		}

	PORTC |= (1<<_CS);			// CS#
	}

/* ---------------------------------------------

	BoẍobNOEhFŖ߂

--------------------------------------------- */
void OLEDFillBox(WORD x, WORD y, WORD dx, WORD dy, WORD color)
	{
	DWORD	loop;
	
	loop= (DWORD)dx * (DWORD)dy;
	
	SetWindowParam	(x, y, x+dx-1, y+dy-1);
	
	OLEDFill		(loop, color);
	}

void OLEDFillBoxA(void)
	{
	OLEDFillBox	(attr.gx, attr.gy, attr.gdx, attr.gdy, attr.bgcolor);
	}

/* ---------------------------------------------

	LN^W̃Zbg(Locate)

--------------------------------------------- */
void OLEDSetLocate(WORD x, WORD y)
	{
	attr.chrx= x;
	attr.chry= y;
	}

/* ---------------------------------------------

	1\
	chr <-- LN^R[h
	ASCII͈͊O16i2\s

--------------------------------------------- */
void OLEDPutChr(char chr)
	{
	const prog_uchar *p1, *p2;

	/* \̔ƃtHg؂ւ */
	if	(' ' <= chr && chr <= 0x7f)
		{				// ʏASCIItHg
		chr -= ' ';
		p1= ankfont + (chr << 3);
		p2= 0;
		}
	else
		{				// \͈͊O̕16i\
		p1= exfont + ((chr & 0x0f) << 3);
		p2= exfont + ((chr & 0xf0) >> 1);
		}
	
	OLEDPutChrP	(p1, p2);		// \s
	}

/* tHgf[^ROMAhXw肵ĕ\Tu[` */
void OLEDPutChrP(const prog_uchar *p1, const prog_uchar *p2)
	{
	WORD		x, y;
	BYTE		c1, l1, bit;
	
	x= attr.chrx;	y= attr.chry;		// \EBhE̐ݒ
	SetWindowParam	(x, y, x+7, y+7);

	/* Data */
	PORTC &= ~(1<<_CS);			// CS#

	/* 8*8dotLN^WJ` */
	for	(l1=0; l1<8; l1++)
		{				// tHgf[^ǂݏo
		c1= pgm_read_byte_near(p1++);	// tHg
		if	(p2)			// IvV(dˍ킹)tHg
			{ c1 |= pgm_read_byte_near(p2++) << 4; }

		/* 8dotWJ` */
		for	(bit=0x80; bit; bit>>=1)
			{
			if	(c1 & bit)		// Dot=ON: tHAOEhFŕ`
				{
				PORTD= (BYTE)(attr.fgcolor >> 8);
				PORTC &= ~(1<<_WR);	// WR#
				PORTC |= (1<<_WR);	// WR#
				PORTD= (BYTE)(attr.fgcolor & 0xff);
				}
			else				// Dot=OFF: obNOEhFŕ`
				{
				PORTD= (BYTE)(attr.bgcolor >> 8);
				PORTC &= ~(1<<_WR);	// WR#
				PORTC |= (1<<_WR);	// WR#
				PORTD= (BYTE)(attr.bgcolor & 0xff);
				}
			
			PORTC &= ~(1<<_WR);		// WR#
			PORTC |= (1<<_WR);		// WR#
			}
		}
	
	PORTC |= (1<<_CS);				// CS#

	/* LN^WE8doti߂ */
	x += 8;
	if	(x > (SCREENGX - 8))
		{
		x= 0;	y += 8;
		
		if	(y > (SCREENGY - 8))
			y= 0;
		}
	attr.chrx= x;	attr.chry= y;
	}

/* ---------------------------------------------

	\
	str <-- ̃|C^(ROMAhX)

--------------------------------------------- */
void OLEDPutStr(const prog_char *str)
	{
	WORD	r, g, b;
	BYTE	c;

	while	((c= pgm_read_byte_near(str++)) != 0)
		{
		/* @\s̔ */
		if	(c != GC_ESCAPE)
			{
			OLEDPutChr(c);		// ʏ̕
			continue;
			}
		
		/* @\̏ */
		switch	(c= pgm_read_byte_near(str++))
			{
			case	GC_LOCATE:
				attr.chrx  = pgm_read_byte_near(str++) * 256;
				attr.chrx += pgm_read_byte_near(str++);
				attr.chry  = pgm_read_byte_near(str++);
				break;

			case	GC_BGCOLOR:
				r= pgm_read_byte_near(str++);
				g= pgm_read_byte_near(str++);
				b= pgm_read_byte_near(str++);
				attr.bgcolor= ((b & 0xf8) << 8) | ((g & 0xfc) << 3) | ((r & 0xf8) >> 3);
				break;

			case	GC_FGCOLOR:
				r= pgm_read_byte_near(str++);
				g= pgm_read_byte_near(str++);
				b= pgm_read_byte_near(str++);
				attr.fgcolor= ((b & 0xf8) << 8) | ((g & 0xfc) << 3) | ((r & 0xf8) >> 3);
				break;
			
			case	GC_BOXF:
				attr.gx  = pgm_read_byte_near(str++) * 256;
				attr.gx += pgm_read_byte_near(str++);
				attr.gy  = pgm_read_byte_near(str++);
				attr.gdx = pgm_read_byte_near(str++) * 256;
				attr.gdx+= pgm_read_byte_near(str++);
				attr.gdy = pgm_read_byte_near(str++);
				OLEDFillBoxA();
				break;
			
			case	GC_IMGCHR:	// R}hȌ8Byte𕶎tHgƂĕ\
				OLEDPutChrP	((const prog_uchar *)str, 0);
				str += 8;
				break;
			}
		}
	}


/* ---------------------------------------------

	w蒷̕\
	str <-- ̃|C^(RAMAhX)

--------------------------------------------- */
void OLEDFixWStr(char *str, WORD width)
	{
	WORD	i;
	
	for	(i=0; i<width; i++)
		{
		if	(*str)
			{ OLEDPutChr(*str++); continue; }
		while	(i++ < width)	OLEDPutChr(' ');
		break;
		}
	}

